import bionty as bt


class Species(bt.Species):
    """Bionty Species.

    See `bionty.Species <https://lamin.ai/docs/bionty/bionty.Species>`__.
    """

    def __init__(self, id="common_name") -> None:
        super().__init__(id=id)


class Gene(bt.Gene):
    """Bionty Gene.

    See `bionty.Gene <https://lamin.ai/docs/bionty/bionty.Gene>`__.
    """

    def __init__(self, species="human", id="ensembl_gene_id") -> None:
        super().__init__(species=species, id=id)


class Protein(bt.Protein):
    """Bionty Protein.

    See `bionty.Protein <https://lamin.ai/docs/bionty/bionty.Protein>`__.
    """

    def __init__(self, species="human", id="uniprotkb_id") -> None:
        super().__init__(species=species, id=id)


class CellMarker(bt.CellMarker):
    """Bionty CellMarker.

    See `bionty.CellMarker <https://lamin.ai/docs/bionty/bionty.CellMarker>`__.
    """

    def __init__(self, species="human", id="name") -> None:
        super().__init__(species=species, id=id)


class CellType(bt.CellType):
    """Bionty CellType.

    See `bionty.CellType <https://lamin.ai/docs/bionty/bionty.CellType>`__.
    """

    def __init__(self, id="ontology_id") -> None:
        super().__init__(id=id)


class Tissue(bt.Tissue):
    """Bionty Tissue.

    See `bionty.Tissue <https://lamin.ai/docs/bionty/bionty.Tissue>`__.
    """

    def __init__(self, id="ontology_id") -> None:
        super().__init__(id=id)


class Disease(bt.Disease):
    """Bionty Disease.

    See `bionty.Disease <https://lamin.ai/docs/bionty/bionty.Disease>`__.
    """

    def __init__(self, id="ontology_id") -> None:
        super().__init__(id=id)



from bionty import lookup as bionty_lookup
from lnbfx import lookup as lnbfx_lookup


class lookup:
    """Lookup identifiers of knowledge tables."""

    pass


def _get_all_methods(lookup):
    return [i for i in dir(lookup) if not i.startswith("_")]


for module in [bionty_lookup, lnbfx_lookup]:
    methods = _get_all_methods(module)
    for method in methods:
        model = getattr(module, method)
        setattr(lookup, method, staticmethod(model))



import base64
import hashlib
from pathlib import Path
from typing import Any, List, Optional, Set, Tuple, Union

import anndata as ad
import lnschema_bionty as bionty
import pandas as pd
from lamin_logger import logger
from lndb_setup import settings
from lnschema_core import DObject as lns_DObject
from lnschema_core import Features, Run, Storage
from typeguard import typechecked

from lamindb.knowledge import CellMarker, Gene, Protein

from .dev._core import get_name_suffix_from_filepath
from .dev.db._add import add
from .dev.db._select import select
from .dev.file import load_to_memory
from .dev.object import infer_suffix, size_adata, write_to_file
from .schema._table import table_meta


def serialize(
    data: Union[Path, str, pd.DataFrame, ad.AnnData], name, format
) -> Tuple[Any, Path, str, str]:
    """Serialize a data object that's provided as file or in memory."""
    memory_rep = None
    if isinstance(data, (Path, str)):
        local_filepath = Path(data)
        name, suffix = get_name_suffix_from_filepath(local_filepath)
    elif isinstance(data, (pd.DataFrame, ad.AnnData)):
        if name is None:
            raise RuntimeError("Provide name if recording in-memory data.")
        memory_rep = data
        suffix = infer_suffix(data, format)
        local_filepath = Path(f"{name}{suffix}")
        if suffix != ".zarr":
            write_to_file(data, local_filepath)
    else:
        raise NotImplementedError("Recording not yet implemented for this type.")
    return memory_rep, local_filepath, name, suffix


def get_hash(local_filepath, suffix):
    if suffix != ".zarr":  # if not streamed
        hash = hash_file(local_filepath)
        result = select(lns_DObject, hash=hash).one_or_none()
        if result is not None:
            logger.warning(
                "Based on the MD5 hash, the same data object is already"
                f" in the DB: {result}"
            )
    else:
        hash = None
    return hash


def get_features_records(
    parsing_id: str,
    features_ref: Union[Gene, Protein, CellMarker],
    df_curated: pd.DataFrame,
) -> List[Union[Gene, Protein, CellMarker]]:
    # insert species entry if not exists
    species = select(bionty.Species, common_name=features_ref.species).one_or_none()
    if species is None:
        species = add(bionty.Species(common_name=features_ref.species))

    model = table_meta.get_model(f"bionty.{features_ref.entity}")

    # all existing feature records of the species in the db
    stmt = (
        select(model)
        .where(getattr(model, parsing_id).in_(df_curated.index))
        .where(getattr(model, "species_id") == species.id)
    )
    records = stmt.all()
    records_df = df_curated.index.intersection(stmt.df()[parsing_id])

    # new records to be appended
    new_ids = df_curated.index.difference(records_df)
    if len(new_ids) > 0:
        # mapped new_ids
        mapped = features_ref.df.loc[features_ref.df.index.intersection(new_ids)].copy()
        mapped.index.name = parsing_id
        if mapped.shape[0] > 0:
            for kwargs in mapped.reset_index().to_dict(orient="records"):
                kwargs["species_id"] = species.id
                record = model(**kwargs)
                records.append(record)
        # unmapped new_ids
        unmapped = set(new_ids).difference(mapped.index)
        if len(unmapped) > 0:
            for i in unmapped:
                record = model(**{parsing_id: i, "species_id": species.id})
                records.append(record)

    return records


def parse_features(
    df: pd.DataFrame, features_ref: Union[CellMarker, Gene, Protein]
) -> None:
    """Link features to a knowledge table.

    Args:
        df: a DataFrame
        features_ref: Features reference class.
    """
    parsing_id = features_ref._id_field

    # Add and curate features against a knowledge table
    column = None
    if parsing_id in df.columns:
        column = parsing_id
    else:
        logger.warning(f"{parsing_id} column not found, using index as features.")
    df_curated = features_ref.curate(df=df, column=column)

    # logging of curation
    n = df_curated["__curated__"].count()
    n_mapped = df_curated["__curated__"].sum()
    log = {  # noqa  TODO: store this somewhere in the db
        "feature": parsing_id,
        "n_mapped": n_mapped,
        "percent_mapped": round(n_mapped / n * 100, 1),
        "unmapped": df_curated.index[~df_curated["__curated__"]],
    }

    features_hash = hash_set(set(df_curated.index))

    features = select(
        Features,
        id=features_hash,
        type=features_ref.entity,
    ).one_or_none()
    if features is not None:
        return features  # features already exists!

    features = Features(id=features_hash, type=features_ref.entity)
    records = get_features_records(parsing_id, features_ref, df_curated)

    if isinstance(features_ref, Gene):
        for record in records:
            features.genes.append(record)
    elif isinstance(features_ref, Protein):
        for record in records:
            features.proteins.append(record)
    elif isinstance(features_ref, CellMarker):
        for record in records:
            features.cell_markers.append(record)

    return features


def get_features(dobject_privates, features_ref):
    """Updates dobject in place."""
    memory_rep = dobject_privates["_memory_rep"]
    if memory_rep is None:
        memory_rep = load_to_memory(dobject_privates["_local_filepath"])
    try:
        df = getattr(memory_rep, "var")  # for AnnData objects
        if callable(df):
            df = memory_rep
    except AttributeError:
        df = memory_rep
    return parse_features(df, features_ref)


def get_run(run: Optional[Run]) -> Run:
    if run is None:
        from ._nb import _run

        run = _run
        if run is None:
            raise ValueError("Pass a Run record.")
    return run


@typechecked
def get_dobject_kwargs_from_data(
    data: Union[Path, str, pd.DataFrame, ad.AnnData],
    *,
    name: Optional[str] = None,
    features_ref: Optional[Union[CellMarker, Gene, Protein]] = None,
    source: Optional[Run] = None,
    format: Optional[str] = None,
):
    """Record a data object.

    Guide: :doc:`/db/guide/ingest`.

    Args:
        data: Filepath or in-memory data.
        name: Name of the data object, required if an in-memory object is passed.
        features_ref: Reference against which to link features.
        source: The data transform that links to the data source of the data object.
        id: The id of the dobject.
        format: Whether to use `h5ad` or `zarr` to store an `AnnData` object.
    """
    run = get_run(source)
    memory_rep, local_filepath, name, suffix = serialize(data, name, format)
    if suffix != ".zarr":
        size = Path(local_filepath).stat().st_size
    else:
        size = size_adata(memory_rep)
    hash = get_hash(local_filepath, suffix)
    storage = select(Storage, root=str(settings.instance.storage_root)).one()
    dobject_privates = dict(
        _local_filepath=local_filepath,
        _memory_rep=memory_rep,
    )
    if features_ref is not None:
        features = [get_features(dobject_privates, features_ref)]  # has to be list!
    else:
        features = []
    dobject_kwargs = dict(
        name=name,
        suffix=suffix,
        hash=hash,
        run_id=run.id,
        size=size,
        storage_id=storage.id,
        source=run,
        features=features,
    )
    return dobject_kwargs, dobject_privates


def to_b64_str(bstr: bytes):
    b64 = base64.urlsafe_b64encode(bstr).decode().strip("=")
    return b64


# a lot to read about this: lamin-notes/2022/hashing
def hash_set(s: Set[str]) -> str:
    bstr = ":".join(sorted(s)).encode("utf-8")
    # as we're truncating at 20 b64, we choose md5 over sha512
    return to_b64_str(hashlib.md5(bstr).digest())[:20]


def hash_file(path: Path) -> str:
    # based on https://stackoverflow.com/questions/3431825/generating-an-md5-hash-of-a-file  # noqa
    hash_md5 = hashlib.md5()
    with open(path, "rb") as file:
        for chunk in iter(lambda: file.read(4096), b""):
            hash_md5.update(chunk)
    return to_b64_str(hash_md5.digest())



"""Access the schema.

- The core entities are the basis for tracking any data and are available from
  `lns.<entity>`.
- Additional mounted schema modules provide domain-specific entities and are
  available via `lns.<module>.<entity>`.

Import this submodule as::

   import lamindb.schema as lns

Core entities
=============

Data objects ({class}`lamindb.DObject`) are transformed by runs:

.. autosummary::
   :toctree: .

   Run

Runs transform data using code:

.. autosummary::
   :toctree: .

   Pipeline
   Notebook

Grouping data objects as sets or by features:

.. autosummary::
   :toctree: .

   DSet
   Features

Users, projects, storage locations, and usage:

.. autosummary::
   :toctree: .

   User
   Project
   Storage
   Usage

See the source code `here <https://lamin.ai/docs/lnschema-core>`__.

Exemplary extensions
====================

Any LaminDB schema module that has been mounted to an instance can be accessed like the bionty, wetlab, bfx modules below:

- `bionty <https://lamin.ai/docs/lnschema-bionty/api>`__: Knowledge-managed biological entities.
- `wetlab <https://lamin.ai/docs/lnschema-wetlab/api>`__: Wetlab operations.


Helper tools
============

.. autosummary::
   :toctree: .

   view
   list_tables
   dev

"""
import importlib as _importlib

from lndb_setup import settings as _settings
from lndb_setup._setup_schema import get_schema_module_name as _get_schema_module_name
from lnschema_core import (
    DObject,
    DSet,
    Features,
    Notebook,
    Pipeline,
    Project,
    Run,
    Storage,
    Usage,
    User,
    dev,
)
from packaging import version as _v

_check_v = {
    "bionty": "0.6.2",
    "wetlab": "0.12.1",
    "bfx": "0.7.0",
}

for name in _settings.instance.schema:
    _module = _importlib.import_module(_get_schema_module_name(name))
    if name in _check_v:
        if _v.parse(_module.__version__) != _v.parse(_check_v[name]):
            raise RuntimeError(f"lamindb needs lnschema_{name}=={_check_v[name]}")
    globals()[_module._name] = _module

from ._core import list_tables, view

list_entities = list_tables  # backward compat



from typing import List, Union, overload  # noqa

import sqlmodel as sqm
from lndb_setup import settings
from lnschema_core import DObject, Usage
from lnschema_core.link import RunIn

from ._logger import colors, logger
from .dev._core import storage_key_from_dobject
from .dev.db._select import select
from .dev.file import delete_storage


@overload
def delete(record: sqm.SQLModel) -> None:
    ...


@overload
def delete(records: List[sqm.SQLModel]) -> None:  # type: ignore
    ...


@overload
def delete(entity: sqm.SQLModel, **fields) -> None:  # type: ignore
    ...


def delete(  # type: ignore
    record: Union[sqm.SQLModel, List[sqm.SQLModel]], **fields
) -> None:
    """Delete data records & data objects.

    Guide: :doc:`/db/guide/add-delete`.

    Example:

    >>> # Delete by record
    >>> experiment = ln.select(Experiment, id=experiment_id)
    >>> ln.delete(experiment)
    >>> # Delete data objects
    >>> dobject = ln.select(DObject, id=dobject_id)
    >>> ln.delete(dobject)
    >>> # Delete by fields
    >>> ln.delete(DObject, id=dobject_id)

    Args:
        record: One or multiple records as instances of `SQLModel`.
    """
    if isinstance(record, list):
        records = record
    elif isinstance(record, sqm.SQLModel):
        records = [record]
    else:
        model = record
        results = select(model, **fields).all()
        if len(results) == 0:
            return None
        records = results
    session = settings.instance.session()
    for record in records:
        if isinstance(record, DObject):
            # delete usage events related to the dobject that's to be deleted
            events = session.exec(
                sqm.select(Usage).where(Usage.dobject_id == record.id)
            )
            for event in events:
                session.delete(event)
            session.commit()
            # delete run_ins related to the dobject that's to be deleted
            run_ins = session.exec(
                sqm.select(RunIn).where(RunIn.dobject_id == record.id)
            )
            for run_in in run_ins:
                session.delete(run_in)
            session.commit()
        session.delete(record)
        session.commit()
        settings.instance._update_cloud_sqlite_file()
        logger.success(
            f"Deleted {colors.yellow(f'row {record}')} in"
            f" {colors.blue(f'table {type(record).__name__}')}."
        )
        if settings.instance._session is None:
            session.close()
        if isinstance(record, DObject):
            # TODO: do not track deletes until we come up
            # with a good design that respects integrity
            # track_usage(entry.id, "delete")
            storage_key = storage_key_from_dobject(record)
            delete_storage(storage_key)
            logger.success(
                f"Deleted {colors.yellow(f'object {storage_key}')} from storage."
            )



"""Small example datasets.

.. autosummary::
   :toctree: .

   file_fcs
   file_jpg_paradisi05
   dir_scrnaseq_cellranger
   anndata_mouse_sc_lymph_node
   anndata_human_immune_cells
"""

from ._core import (
    anndata_human_immune_cells,
    anndata_mouse_sc_lymph_node,
    dir_scrnaseq_cellranger,
    file_fcs,
    file_jpg_paradisi05,
)



"""Helpers for the db API the database.

Select:

.. autosummary::
   :toctree: .

   SelectStmt
   ExecStmt

Other:

.. autosummary::
   :toctree: .

   session
"""

from ._add import add
from ._core import session
from ._select import ExecStmt, SelectStmt, select



"""Manage Jupyter notebooks.

.. autosummary::
   :toctree: .

   header
   publish

For more background, see `nbproject <https://lamin.ai/docs/nbproject>`__.
"""
from ._nb import header, publish  # noqa



"""Developer API.

.. autosummary::
   :toctree: .

   db
   datasets
   file
   object

Utilities:

.. autosummary::
   :toctree: .

   doc_args
"""

from lnschema_core import id  # noqa

from . import datasets  # noqa
from . import db  # noqa
from . import file, object  # noqa
from ._docs import doc_args  # noqa



import lnschema_core as core
from lamin_logger import logger
from lndb_setup import settings

from .dev._core import filepath_from_dobject
from .dev.db._track_usage import track_usage
from .dev.file import load_to_memory


def populate_runin(dobject: core.DObject, run: core.Run):
    session = settings.instance.session()
    result = session.get(core.link.RunIn, (run.id, dobject.id))
    if result is None:
        session.add(
            core.link.RunIn(
                run_id=run.id,
                dobject_id=dobject.id,
            )
        )
        session.commit()
        logger.info(f"Added dobject ({dobject.id}) as input for run ({run.id}).")
        settings.instance._update_cloud_sqlite_file()
    if settings.instance._session is None:
        session.close()


def load(dobject: core.DObject, stream: bool = False):
    """Load data object.

    Returns object associated with the stored `dobject`.

    Populates `RunIn` when called from a notebook.

    Guide: :doc:`/db/guide/select-load`.
    """
    if stream and dobject.suffix not in (".h5ad", ".zarr"):
        logger.warning(f"Ignoring stream option for a {dobject.suffix} object.")

    filepath = filepath_from_dobject(dobject)
    from lamindb._nb import _run as nb_run

    if nb_run is None:
        logger.warning(
            "Input tracking for runs through `load` is currently only implemented for"
            " notebooks."
        )
    else:
        populate_runin(dobject, nb_run)
    track_usage(dobject.id, "load")
    return load_to_memory(filepath, stream=stream)



from pathlib import Path
from urllib.request import urlretrieve

import anndata as ad


def file_fcs() -> Path:
    """Return fcs file example."""
    filepath, _ = urlretrieve(
        "https://lamindb-test.s3.amazonaws.com/example.fcs", "example.fcs"
    )
    return Path(filepath)


def file_jpg_paradisi05() -> Path:
    """Return jpg file example.

    Originally from: https://upload.wikimedia.org/wikipedia/commons/2/28/Laminopathic_nuclei.jpg
    """  # noqa
    filepath, _ = urlretrieve(
        "https://lamindb-test.s3.amazonaws.com/Laminopathic_nuclei.jpg",
        "paradisi05_laminopathic_nuclei.jpg",
    )
    return Path(filepath)


def dir_scrnaseq_cellranger() -> Path:
    """Directory with exemplary scrnaseq cellranger output."""
    filepath, _ = urlretrieve(
        "https://lamindb-test.s3.amazonaws.com/cellranger_run_001.zip"
    )
    from zipfile import ZipFile

    with ZipFile(filepath, "r") as zipObj:
        # Extract all the contents of zip file in current directory
        zipObj.extractall(path=".")

    return Path("cellranger_run_001")


def anndata_mouse_sc_lymph_node() -> ad.AnnData:
    """Mouse lymph node scRNA-seq dataset from EBI.

    Subsampled to 10k genes.

    From: https://www.ebi.ac.uk/arrayexpress/experiments/E-MTAB-8414/
    """
    filepath, _ = urlretrieve("https://lamindb-test.s3.amazonaws.com/E-MTAB-8414.h5ad")
    return ad.read(filepath)


def anndata_human_immune_cells() -> ad.AnnData:
    """Cross-tissue immune cell analysis reveals tissue-specific features in humans.

    From: https://cellxgene.cziscience.com/collections/62ef75e4-cbea-454e-a0ce-998ec40223d3  # noqa
    Dataset: Global

    To reproduce the subsample:
    >>> adata = sc.read('Global.h5ad')
    >>> adata.obs = adata.obs[['donor_id', 'tissue', 'cell_type', 'assay', 'tissue_ontology_term_id', 'cell_type_ontology_term_id', 'assay_ontology_term_id']].copy() . # noqa
    >>> sc.pp.subsample(adata, fraction=0.005)
    >>> del adata.uns["development_stage_ontology_term_id_colors"]
    >>> del adata.uns["sex_ontology_term_id_colors"]
    >>> sc.write('human_immune.h5ad', adata)
    """
    filepath, _ = urlretrieve("https://lamindb-test.s3.amazonaws.com/human_immune.h5ad")
    return ad.read(filepath)


# def schmidt22_crispra_gws_IFNG() -> Path:
#     """CRISPRi screen dataset of Schmidt22.

#     Originally from: https://zenodo.org/record/5784651
#     """
#     filepath, _ = urlretrieve(
#         "https://lamindb-test.s3.amazonaws.com/schmidt22-crispra-gws-IFNG.csv",
#         "schmidt22-crispra-gws-IFNG.csv",
#     )
#     return Path(filepath)


# def schmidt22_perturbseq() -> Path:
#     """Perturb-seq dataset of Schmidt22.

#     Subsampled and converted to h5ad from R file: https://zenodo.org/record/5784651

#     To reproduce the subsample:
#     >>> adata = sc.read('HuTcellsCRISPRaPerturbSeq_Re-stimulated.h5ad')
#     >>> adata.obs = adata.obs[['cluster_name']]
#     >>> del adata.obsp
#     >>> del adata.var['features']
#     >>> del adata.obsm['X_pca']
#     >>> del adata.uns
#     >>> del adata.raw
#     >>> del adata.varm
#     >>> adata.obs = adata.obs.reset_index()
#     >>> del adata.obs['index']
#     >>> sc.pp.subsample(adata, 0.03)
#     >>> adata.write('schmidt22_perturbseq.h5ad')
#     """
#     filepath, _ = urlretrieve(
#         "https://lamindb-test.s3.amazonaws.com/schmidt22_perturbseq.h5ad",
#         "schmidt22_perturbseq.h5ad",
#     )
#     return Path(filepath)


# def dir_scrnaseq_cellranger_schmidt22() -> Path:
#     """BFXoutput directory containing Schmidt22_perturbseq."""
#     filepath, _ = urlretrieve(
#         "https://lamindb-test.s3.amazonaws.com/scrnaseq-cellranger-schmidt22.zip",
#     )
#     from zipfile import ZipFile

#     with ZipFile(filepath, "r") as zipObj:
#         # Extract all the contents of zip file in current directory
#         zipObj.extractall(path=".")

#     return Path("scrnaseq-cellranger-schmidt22")



from functools import partial
from typing import List, Union, overload  # noqa

import sqlmodel as sqm
from lndb_setup import settings
from lnschema_core import DObject

from ..file import store_file, write_adata_zarr
from ..file._file import print_hook
from ._core import dobject_to_sqm
from ._select import select


@overload
def add(record: sqm.SQLModel, use_fsspec: bool = True) -> sqm.SQLModel:
    ...


# Currently seeing the following error without type ignore:
# Overloaded function signature 2 will never be matched: signature 1's parameter
# type(s) are the same or broader
@overload
def add(  # type: ignore
    records: List[sqm.SQLModel], use_fsspec: bool = True
) -> List[sqm.SQLModel]:
    ...


@overload
def add(  # type: ignore
    entity: sqm.SQLModel, use_fsspec: bool = True, **fields
) -> Union[sqm.SQLModel, List[sqm.SQLModel]]:
    ...


def add(  # type: ignore  # no support of different naming of args across overloads
    record: Union[sqm.SQLModel, List[sqm.SQLModel]], use_fsspec: bool = True, **fields
) -> Union[sqm.SQLModel, List[sqm.SQLModel]]:
    """Insert or update data records in the DB ("metadata" entities).

    Inserts a new :term:`record` if the corresponding row doesn't exist.
    Updates the corresponding row with the record if it exists.

    To update a row, query it with `.get` or `.select` and modify it before
    passing it to `add`.

    Guide: :doc:`/db/guide/add-delete`.

    Example:

    >>> # add a record (by passing a record)
    >>> ln.add(wetlab.Experiment(name="My test", biometa_id=test_id))
    >>> # update an existing record
    >>> experiment = ln.select(wetlab.Experiment, id=experiment_id).one()
    >>> experiment.name = "New name"
    >>> ln.add(experiment)
    >>> # add a record by fields if not yet exists
    >>> ln.add(wetlab.Experiment, name="My test", biometa_id=test_id)

    Args:
        record: One or multiple records as instances of `SQLModel`.
        use_fsspec: Whether to use fsspec.
    """
    if isinstance(record, list):
        records = record
    elif isinstance(record, sqm.SQLModel):
        records = [record]
    else:
        model = dobject_to_sqm(record)
        results = select(model, **fields).all()
        if len(results) == 1:
            return results[0]
        elif len(results) > 1:
            return results
        else:
            records = [model(**fields)]
    for record in records:
        if isinstance(record, DObject) and hasattr(record, "_local_filepath"):
            upload_data_object(record, use_fsspec=use_fsspec)
    session = settings.instance.session()
    for record in records:
        session.add(record)
    session.commit()
    for record in records:
        session.refresh(record)
    if settings.instance._session is None:
        session.close()
    settings.instance._update_cloud_sqlite_file()
    if len(records) > 1:
        return records
    else:
        return records[0]


def upload_data_object(dobject, use_fsspec: bool = True) -> None:
    """Store and add dobject and its linked entries."""
    dobject_storage_key = f"{dobject.id}{dobject.suffix}"

    if dobject.suffix != ".zarr":
        store_file(dobject._local_filepath, dobject_storage_key, use_fsspec=use_fsspec)
    else:
        storagepath = settings.instance.storage.key_to_filepath(dobject_storage_key)
        print_progress = partial(print_hook, filepath=dobject._local_filepath)
        write_adata_zarr(dobject._memory_rep, storagepath, callback=print_progress)



from typing import List, Union

import nbproject as nb
from lamin_logger import logger
from lnschema_core import Notebook, Run

_notebook: Notebook = None  # Notebook of this Python session
_run: Run = None  # run of this Python session


def header(
    *,
    run: Union[str, None] = None,
    pypackage: Union[str, List[str], None] = None,
    filepath: Union[str, None] = None,
    env: Union[str, None] = None,
):
    """Display metadata and start tracking dependencies.

    If the notebook has no nbproject metadata, initializes & writes metadata to disk.

    Args:
        run: If `None`, loads the latest run of the notebook (or creates on if
            there isn't any).

            If `"new"`, instantiates a new `Run` record.

        pypackage: One or more python packages to track.
        filepath: Filepath of notebook. Only needed if automatic inference fails.
        env: Editor environment. Only needed if automatic inference fails.
            Pass `'lab'` for jupyter lab and `'notebook'` for jupyter notebook,
            this can help to identify the correct mechanism for interactivity
            when automatic inference fails.
    """
    nb.header(pypackage=pypackage, filepath=filepath, env=env)

    import lamindb as ln
    import lamindb.schema as lns

    notebook = ln.select(
        lns.Notebook, id=nb.meta.store.id, v=nb.meta.store.version
    ).one_or_none()
    if notebook is None:
        notebook = lns.Notebook(
            id=nb.meta.store.id, v=nb.meta.store.version, name=nb.meta.live.title
        )
        notebook = ln.add(notebook)
        logger.info(f"Added notebook: {notebook.id} v{notebook.v}")

    # at this point, we have a notebook object
    global _notebook
    _notebook = notebook

    # check user input
    # if isinstance(run, lns.Run):
    # This here might be something we may want in the future
    # but checking all the cases in which that run record has integrity
    # is quite a bit of code - not now!
    #     run_test = ln.select(lns.Run, id=run.id).one_or_none()
    #     if run_test is None:
    #         logger.info("Passed run does not exist, adding it")
    #         ln.add(run)
    if run is None:
        # retrieve the latest run
        run = (
            ln.select(lns.Run, notebook_id=notebook.id, notebook_v=notebook.v)
            .order_by(lns.Run.created_at.desc())
            .first()
        )
        if run is not None:
            logger.info(f"Loaded run: {run.id}")  # type: ignore
    elif run != "new":
        raise ValueError("Pass run='new' to header()!")

    # create a new run if doesn't exist yet or is requested by the user ("new")
    if run is None or run == "new":
        run = lns.Run(notebook_id=notebook.id, notebook_v=notebook.v)
        run = ln.add(run)
        logger.info(f"Created run: {run.id}")  # type: ignore

    # at this point, we have a run object
    global _run
    _run = run


def publish(version: str = None, i_confirm_i_saved: bool = False):
    """Publish the notebook.

    Args:
        version: Notebook version to publish.
        i_confirm_i_saved: Only relevant outside Jupyter Lab & Notebook as a
            safeguard against losing the editor buffer content.
    """
    from nbproject._publish import finalize_publish, run_checks_for_publish

    import lamindb as ln
    import lamindb.schema as lns

    result = run_checks_for_publish(
        calling_statement="publish(", i_confirm_i_saved=i_confirm_i_saved
    )
    if result != "checks-passed":
        return result
    finalize_publish(calling_statement="publish(", version=version)
    # update DB
    notebook = ln.select(Notebook, id=_notebook.id, v=_notebook.v).one()
    # update version
    notebook.name = nb.meta.live.title
    if version != _notebook.v:
        notebook_add = lns.Notebook(id=notebook.id, v=version, name=notebook.name)
    else:
        notebook_add = notebook
    ln.add(notebook_add)
    if version != _notebook.v:
        _run.notebook_v = version
        ln.add(_run)
        ln.delete(notebook)



import anndata
import fsspec
from anndata import AnnData
from lndb_setup import settings


def h5ad_to_anndata(filekey) -> AnnData:
    """h5ad → AnnData."""
    return anndata.read(settings.instance.storage.local_filepath(filekey))


def read_adata_h5ad(filepath, **kwargs) -> AnnData:
    if not isinstance(filepath, str):
        filepath = str(filepath)
    with fsspec.open(filepath, mode="rb") as file:
        adata = anndata.read_h5ad(file, backed=False, **kwargs)
        return adata



from bionty import __version__ as bionty_v
from lndb_setup import __version__ as lndb_setup_v
from lnschema_core import __version__ as lnschema_core_v
from nbproject import __version__ as nbproject_v
from packaging import version

if version.parse(lndb_setup_v) != version.parse("0.25.0"):
    raise RuntimeError("lamindb needs lndb_setup==0.25.0")

if version.parse(lnschema_core_v) != version.parse("0.23.1"):
    raise RuntimeError("lamindb needs lnschema_core==0.23.1")

if version.parse(bionty_v) != version.parse("0.5.7"):
    raise RuntimeError("lamindb needs bionty==0.5.7")

if version.parse(nbproject_v) < version.parse("0.8.0"):
    raise RuntimeError("lamindb needs nbproject>=0.8.0")



from typing import Tuple, Union

import sqlmodel as sqm
from lndb_setup import settings
from lnschema_core import DObject


def session() -> sqm.Session:
    """Get connection session to DB engine.

    Returns a `sqlmodel.Session` object.
    """
    return settings.instance.session()


def dobject_to_sqm(entity: Union[sqm.SQLModel, Tuple[sqm.SQLModel]]):
    def if_dobject(entity):
        if entity.__class__.__name__ == "type" and entity.__name__ == "DObject":
            return DObject
        else:
            return entity

    if isinstance(entity, tuple):
        entities = list(entity)
        for i, ent in enumerate(entities):
            entities[i] = if_dobject(ent)
        return entities
    else:
        return if_dobject(entity)



"""Storage → object.

File → memory:

.. autosummary::
   :toctree: .

   h5ad_to_anndata

Store files:

.. autosummary::
   :toctree: .

   store_file
   store_png
   delete_storage
"""

from ._file import delete_storage, load_to_memory, store_file
from ._h5ad import h5ad_to_anndata
from ._images import store_png
from ._zarr import read_adata_zarr, write_adata_zarr



"""Knowledge.

Feature tables from bionty:

.. autosummary::
   :toctree: .

   Species
   Gene
   Protein
   CellMarker
   Tissue
   CellType
   Disease

Lookup knowledge table identifiers:

.. autosummary::
   :toctree: .

   lookup
"""


from ._core import CellMarker, CellType, Disease, Gene, Protein, Species, Tissue
from ._lookup import lookup



from lndb_setup import settings
from lnschema_core import Usage
from lnschema_core.dev.type import usage as UsageType


def track_usage(dobject_id, usage_type: UsageType):
    from ._add import add

    usage = Usage(
        type=usage_type,
        user_id=settings.user.id,
        dobject_id=dobject_id,
    )
    add(usage)

    return usage.id



from pathlib import Path

import nox
from lndb_setup.test.nox import build_docs, login_testuser1, run_pre_commit, run_pytest

nox.options.reuse_existing_virtualenvs = True


@nox.session(python=["3.7", "3.8", "3.9", "3.10", "3.11"])
def lint(session: nox.Session) -> None:
    run_pre_commit(session)


@nox.session(python=["3.7", "3.8", "3.9", "3.10", "3.11"])
def build(session):
    login_testuser1(session)
    login_user_2 = "lndb login testuser2@lamin.ai --password goeoNJKE61ygbz1vhaCVynGERaRrlviPBVQsjkhz"  # noqa
    session.run(*(login_user_2.split(" ")), external=True)
    session.install(".[dev,test]")
    instance_dirs = [
        d for d in ["./docs/guide/mydata", "./mydata-test-db"] if Path(d).exists()
    ]
    for instance_dir in instance_dirs:
        clean_instance = f"rm -r {instance_dir}"
        session.run(*clean_instance.split(" "))
    test_db = "lndb init --storage mydata-test-db"
    session.run(*test_db.split(" "), external=True)
    run_pytest(session)
    build_docs(session)



"""LaminDB: Manage data & analyses.

Import the package::

   import lamindb as ln
   import lamindb.schema as lns

Query & inspect data:

.. autosummary::
   :toctree: .

   select
   view

Manipulate data:

.. autosummary::
   :toctree: .

   DObject
   add
   subset
   delete

Schema - entities and their relations:

.. autosummary::
   :toctree: .

   schema

Knowledge management:

.. autosummary::
   :toctree: .

   knowledge

Track Jupyter notebooks:

.. autosummary::
   :toctree: .

   nb

Developer API:

.. autosummary::
   :toctree: .

   session
   settings
   dev

"""

__version__ = "0.21.1"

import warnings

from lndb_setup import settings  # noqa
from lndb_setup._migrate import check_migrate as _check_migrate

from . import _check_versions  # executes checks during import

if settings.instance.storage_root is None:
    raise RuntimeError("Please run `lndb init` to configure an instance.")
_check_migrate(usettings=settings.user, isettings=settings.instance)

from lnschema_core import DObject  # noqa

from . import dev  # noqa
from . import knowledge  # noqa
from . import nb  # noqa
from . import schema  # noqa
from ._delete import delete  # noqa
from ._subset import subset
from ._view import view  # noqa
from .dev.db import session  # noqa
from .dev.db._add import add  # noqa
from .dev.db._select import select  # noqa
from .dev.object._lazy_field import lazy

settings.__doc__ = """Settings.

This re-exports `lndb_setup.settings <https://lamin.ai/docs/lndb-setup/lndb_setup.settings>`__.
"""
